/*
 * blowfish.h:  Header file for blowfish.c
 * Copyright (C) 1997 by Paul Kocher
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.

See blowfish.c for more information about this file.
*/


typedef struct {
  guint32 P[16 + 2];
  guint32 S[4][256];
} BLOWFISH_CTX;

void Blowfish_Init(BLOWFISH_CTX *ctx, unsigned char *key, int keyLen);
void Blowfish_Encrypt(BLOWFISH_CTX *ctx, guint32 *xl, guint32 *xr);
void Blowfish_Decrypt(BLOWFISH_CTX *ctx, guint32 *xl, guint32 *xr);



